#ifndef CEX_CONTAINERS_linked_list_t_H
#define CEX_CONTAINERS_linked_list_t_H

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

typedef int64_t index_t;

typedef struct linked_list_node_t {
  struct linked_list_node_t* prev;
  struct linked_list_node_t* next;

  void* data;
} linked_list_node_t;

typedef struct {
  linked_list_node_t* first;
  index_t length;
  size_t data_size;
} linked_list_t;

typedef void (*list_iterator_t)(index_t index, void* data, void* arg);
typedef bool (*list_filter_t)(index_t index, const void* data, void* arg);

linked_list_t LL_create(size_t data_size);
linked_list_t LL_from_array(size_t data_size, void* array, index_t length);
linked_list_t LL_copy(linked_list_t* list);

void* LL_to_array(linked_list_t* list);

void LL_destroy(linked_list_t* list);

void* LL_insert(linked_list_t* list, index_t index);
void* LL_insert_value(linked_list_t* list, index_t index, const void* data);

void* LL_add(linked_list_t* list);
void* LL_add_value(linked_list_t* list, const void* data);

void LL_remove(linked_list_t* list, index_t index);

void* LL_get(linked_list_t* list, index_t index);
void LL_set(linked_list_t* list, index_t index, const void* data);

void LL_iterate(linked_list_t* list, list_iterator_t func, void* arg);
void LL_iterate_range(linked_list_t* list, list_iterator_t func, void* arg, index_t from, index_t to);

void LL_ITER_COPY(index_t, void* data, void* list);
void LL_ITER_SUM(index_t, void* data, void* result);
void LL_ITER_PRINT(index_t i, void* data, void*);

linked_list_t LL_filter(linked_list_t* list, list_filter_t func, void* arg);
linked_list_t LL_filter_range(linked_list_t* list, list_filter_t func, void* arg, index_t from, index_t to);

#endif